/*
 * Clock and power management.
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "xburst_types.h"
#include "cpm.h"
#include "board.h"

static uint32_t cpm_ctrl_get(void *cpm_base, uint32_t reg)
{
	return REG32(cpm_base + reg);
}

static void cpm_ctrl_set(void *cpm_base, uint32_t reg, uint32_t value)
{
	REG32(cpm_base + reg) = value;
}

static int pll_enabled(void *cpm_base)
{
	return cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLEN;
}

static int pll_bypassed(void *cpm_base)
{
	return cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLBP;
}

// Feedback (9-bit) divider.

static uint16_t get_multiplier(void *cpm_base)
{
	return ((cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLM_MASK) >> CPM_CPPCR_PLLM_BIT) + 2;
}

// Input (5-bit) divider.

static uint8_t get_input_divider(void *cpm_base)
{
	return ((cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLN_MASK) >> CPM_CPPCR_PLLN_BIT) + 2;
}

// Output divider.

static uint8_t get_output_divider(void *cpm_base)
{
	uint8_t od[] = {1, 2, 2, 4};
	return od[(cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLOD_MASK) >> CPM_CPPCR_PLLOD_BIT];
}

// General clock divider.

static uint8_t _get_divider(void *cpm_base, uint32_t reg, uint32_t mask, uint8_t shift)
{
	uint8_t cd[] = {1, 2, 3, 4, 6, 8, 12, 16, 24, 32};
        uint8_t d = (cpm_ctrl_get(cpm_base, reg) & mask) >> shift;
	return (d < 10) ? cd[d] : 1;
}

// CPU clock divider.

static uint8_t get_cpu_divider(void *cpm_base)
{
	return _get_divider(cpm_base, CPM_CPCCR, CPM_CPCCR_CDIV_MASK, CPM_CPCCR_CDIV_BIT);
}

// Memory clock divider.

static uint8_t get_memory_divider(void *cpm_base)
{
	return _get_divider(cpm_base, CPM_CPCCR, CPM_CPCCR_MDIV_MASK, CPM_CPCCR_MDIV_BIT);
}

// Clock source divider for MSC, I2S, LCD and USB.

static uint8_t get_source_divider(void *cpm_base)
{
#ifdef CONFIG_CPU_JZ4730
	return 1;
#else
	return cpm_ctrl_get(cpm_base, CPM_CPCCR) & CPM_CPCCR_PCS ? 1 : 2;
#endif
}

// LCD device clock divider.

static void set_lcd_device_divider(void *cpm_base, uint8_t division)
{
	if (division == 0)
		division = 1;
#ifdef CONFIG_CPU_JZ4730
	else if (division > 16)
		division = 16;
#else
	else if (division > 32)
		division = 32;
#endif

	cpm_ctrl_set(cpm_base, CPM_CPCCR,
		(cpm_ctrl_get(cpm_base, CPM_CPCCR) & ~CPM_CPCCR_LDIV_MASK) |
			((division - 1) << CPM_CPCCR_LDIV_BIT));
}

// LCD pixel clock divider.

static void set_lcd_pixel_divider(void *cpm_base, uint16_t division)
{
#ifndef CONFIG_CPU_JZ4730
	if (division == 0)
		division = 1;
	else if (division > 2048)
		division = 2048;

	cpm_ctrl_set(cpm_base, CPM_LPCDR,
		(cpm_ctrl_get(cpm_base, CPM_LPCDR) & ~CPM_LPCDR_PIXDIV_MASK) |
			(division - 1));
#endif
}

static uint32_t get_pll_frequency(void *cpm_base)
{
	// Test for PLL enable and not PLL bypass.

	if (pll_enabled(cpm_base) && !pll_bypassed(cpm_base))
		return (JZ_EXTAL * get_multiplier(cpm_base)) /
			(get_input_divider(cpm_base) * get_output_divider(cpm_base));
	else
		return JZ_EXTAL;
}

// Clock frequency for MSC, I2S, LCD and USB.

static uint32_t get_output_frequency(void *cpm_base)
{
	return get_pll_frequency(cpm_base) / get_source_divider(cpm_base);
}



/* Public functions. */

// Clock frequency for the CPU.

uint32_t jz4740_cpm_get_cpu_frequency(void *cpm_base)
{
	return get_pll_frequency(cpm_base) / get_cpu_divider(cpm_base);
}

// Clock frequency for the memory.

uint32_t jz4740_cpm_get_memory_frequency(void *cpm_base)
{
	return get_pll_frequency(cpm_base) / get_memory_divider(cpm_base);
}

// Set the device and pixel frequencies, indicating the latter and
// providing the device:pixel frequency ratio.

void jz4740_cpm_set_lcd_frequencies(void *cpm_base, uint32_t pclk, uint8_t ratio)
{
	uint32_t out = get_output_frequency(cpm_base), lcd = pclk * ratio;

	set_lcd_pixel_divider(cpm_base, out / pclk);

	// Limit the device frequency to 150MHz.

	if (lcd > 150000000) lcd = 150000000;

	set_lcd_device_divider(cpm_base, out / lcd);
}

// Update the clock output frequency.

void jz4740_cpm_update_output_frequency(void *cpm_base)
{
	cpm_ctrl_set(cpm_base, CPM_CPCCR, cpm_ctrl_get(cpm_base, CPM_CPCCR) | CPM_CPCCR_CE);
}

// General clock functions.

int jz4740_cpm_have_clock(void *cpm_base)
{
	// NOTE: To check.
#ifdef CONFIG_CPU_JZ4730
	return cpm_ctrl_get(cpm_base, CPM_MSCR) != 0;
#else
	return cpm_ctrl_get(cpm_base, CPM_CLKGR) != 0;
#endif
}

int jz4740_cpm_have_pll(void *cpm_base)
{
	return cpm_ctrl_get(cpm_base, CPM_CPPCR) & CPM_CPPCR_PLLS;
}

void jz4740_cpm_start_clock(void *cpm_base)
{
#ifdef CONFIG_CPU_JZ4730
	cpm_start_ost(cpm_base);
	cpm_ctrl_set(cpm_base, CPM_MSCR, cpm_ctrl_get(cpm_base, CPM_MSCR) & ~CPM_MSCR_MSTP_OST);
#else
	cpm_ctrl_set(cpm_base, CPM_CLKGR, cpm_ctrl_get(cpm_base, CPM_CLKGR) & ~CPM_CLKGR_TCU);
#endif
}

// Peripheral clock control.

void jz4740_cpm_start_lcd(void *cpm_base)
{
#ifndef CONFIG_CPU_JZ4730
	cpm_ctrl_set(cpm_base, CPM_CLKGR, cpm_ctrl_get(cpm_base, CPM_CLKGR) & ~CPM_CLKGR_LCD);
#endif
}

void jz4740_cpm_stop_lcd(void *cpm_base)
{
#ifndef CONFIG_CPU_JZ4730
	cpm_ctrl_set(cpm_base, CPM_CLKGR, cpm_ctrl_get(cpm_base, CPM_CLKGR) | CPM_CLKGR_LCD);
#endif
}

// Register access.

uint32_t jz4740_cpm_ctrl_get(void *cpm_base, uint32_t reg)
{
	return cpm_ctrl_get(cpm_base, reg);
}

void jz4740_cpm_ctrl_set(void *cpm_base, uint32_t reg, uint32_t value)
{
	cpm_ctrl_set(cpm_base, reg, value);
}

/* Top-level initialisation. */

void cpm_init()
{
#ifdef CONFIG_CPU_JZ4730
	cpm_ctrl_set((void *) CPM_BASE, CPM_MSCR, 0xffffffff);
#else
	cpm_ctrl_set((void *) CPM_BASE, CPM_CLKGR, 0x7fff);
#endif
}

